#include <stdio.h>
#include <stdlib.h>
#include <memory.h>

#include "globaldefn.h"
#include "stream.h"

extern int fgetc(FILE *);
extern int fputc(int , FILE *);
extern int fflush(FILE *);
extern int fclose(FILE *);

static void clearerr_fn(FILE *stream);
static int feof_fn(FILE *stream);

#define STREAM_INIT_SIZE	1024

int Stream_getc(ComsStream *stream)
{
	/* If stream is closed return ERR */
	if (stream->size == 0)
	{
		fprintf(stderr,"Stream: Reading on closed stream\n");
		return ERR;
	}

	/* If no more characters return ERR */
	if (++(stream->readIndex) == stream->writeIndex)
	{
		return ERR;
	}

	/* Else return the next character and advance the read pointer */
	return (int) stream->buffer[stream->readIndex];
}

int Stream_putc(int c, ComsStream *stream)
{
	/* If buffer is empty, then this is a closed stream, return ERR */
	if (stream->size == 0)
	{
		fprintf(stderr, "Stream: Writing on closed stream\n");
		return ERR;
	}

	/* If buffer is exhausted, try to increase it by 1.5 times */
	if (stream->writeIndex == stream->size)
	{
		/* If space is not available, return ERR */
		if (! (stream->buffer = realloc((void *)stream->buffer, 
			stream->size + stream->size/2)))
		{
			fprintf(stderr, 
				"Stream: Not enough memory for additional buffer: [%d]\n",
				stream->size + stream->size/2);
			return ERR; 		
		}
		else
			stream->size += stream->size/2;
	}

	/* Store the character in the stream and return the character stored */
	return (int) (stream->buffer[stream->writeIndex++] = (char)c);
}

int Stream_flush(ComsStream *stream)
{
	/* Reset the read and write pointers */
	stream->readIndex = -1;
	stream->writeIndex = 0;
	return OK;
}

int Stream_eof(ComsStream *stream)
{
	/* If read pointer and write pointer are same, then EOF has been reached */
	return (stream->readIndex == stream->writeIndex);
}

int Stream_open(ComsStream *stream)
{
	SP_stream *SP_streamPtr;

	/* Initialise the pointers */
	stream->readIndex = -1;
	stream->writeIndex = 0;

	/* Allocate buffer */
	stream->size = STREAM_INIT_SIZE;
	if (! (stream->buffer = (char *)malloc(stream->size)))
	{
		fprintf(stderr,"Stream: Not enough memory for buffer: [%d]\n", 
			stream->size);
		return ERR;
	}

	/* Inform prolog and create a term for it */
	SP_make_stream(stream, 
		Stream_getc ,
		Stream_putc,
		Stream_flush, Stream_eof, NULL, NULL, &SP_streamPtr);

	/* Make the stream ptr visible */
	SP_show_term(&stream->SP_streamPtr);

	/* Store the stream ptr for later reading */
	SP_put_address(&stream->SP_streamPtr, (void *)SP_streamPtr);

	return OK;
}

int Stream_close(ComsStream *stream)
{
	Stream_flush(stream);
	free(stream->buffer);
	stream->buffer = (char *)NULL;
	stream->size = 0;
	SP_hide_term(&stream->SP_streamPtr);
	return OK;
}

char *Stream_gets(char *string, int n, ComsStream *stream)
{
	int i, c;

	c = '\0';
	for (i = 0; i < n-1 && c != '\n'; i++)
	{
		if ((c = Stream_getc(stream)) == ERR)
		{
			if (i == 0)
			{
				return NULL;
			}
			else
			{
				string[i] = '\0';
				return string;
			}
		}
		else
			string[i] = c;
	}

	string[i] = '\0';

	return string;
}

int Stream_puts(char *string, ComsStream *stream)
{
	/* Note: If you do not want prolog's read to block, put a
		whitespace character after the . */
	if (string) /* Make sure that the pointer is not null */
	{
		while (*string)
		{
			if (Stream_putc((int)(*string), stream) == ERR)
			{
				return ERR;
			}
			string++;
		}
	}

	return OK;
}

/* Link a file pointer and prolog stream */
int Stream_link(ComsStream *stream, FILE *fp)
{
	SP_stream *SP_streamPtr;

	/* Initialise the pointers */
	stream->readIndex = -1;
	stream->writeIndex = 0;

	stream->size = 0;

	/* Inform prolog and create a term for it */
	SP_make_stream(fp, 
		fgetc,
		fputc,
		fflush, feof_fn, clearerr_fn, fclose, &SP_streamPtr);

	SP_show_term(&stream->SP_streamPtr);
	SP_put_address(&stream->SP_streamPtr, (void *)SP_streamPtr);

	return OK;
}

/* clearerr and feof are macros. Cannot be passed to makestream.
	so define a wrapper */
static void clearerr_fn(FILE *stream)
{
	clearerr (stream);
}

static int feof_fn(FILE *stream)
{
	return feof(stream);
}
