/******	This file contains the new timings code from the ver1.96 of the
 ***		Loral code. It has been moved into another file to hide the details
 ***		of the timings computation from the router.
 ***/

#include <kqml/kqmlconf.h>

#include <stdio.h>
#include <string.h>
#include <sys/time.h>
#include <malloc.h>

#include <kqml.h>
#include <kqml/kqmllib.h>

/* used for capturing timing information */
typedef struct kmsg_timing {
	char *id;
	int read_time, write_time;
	long in_time, in_millitm;
	long out_time, out_millitm;
	struct kmsg_timing *next_node, *prev_node;
	} kmsg_timing;

static char *timings_file_name = NULL;
static FILE *timings_file;
static kmsg_timing *kmsg_table = NULL;

typedef enum { WAIT, PROCESS } kmsg_types;

static kmsg_timing * find_kmsg_entry (char *id);
static void remove_kmsg_entry (kmsg_timing *node);
static void write_timings_msg (kmsg_types type, kmsg_timing *node);
void set_kmsg_time (char *id, int timings_type, struct timeval *beg_time, 
	struct timeval *end_time);

static kmsg_timing *
find_kmsg_entry (char *id)
{
	kmsg_timing *node;
  
	if (kmsg_table)
	{
		node = kmsg_table;
		do
		{
			if (kqml_stricmp (node->id, id) == 0)
			{
				return (node);
			}
			else
			{
				node = node->next_node;
			}
		}
		while (node != kmsg_table);
  	}

	/* entry doesn't yet exist, so add it */

	node = malloc (sizeof (kmsg_timing));
	if (node == NULL)
	{
		return NULL;
	}
	node->id = strdup (id);
	if (node->id == NULL)
	{
		free (node);
		return NULL;
	}
	node->read_time = 
		node->write_time = 
		node->in_time = 
		node->in_millitm =
		node->out_time =
		node->out_millitm = 0;

	if (kmsg_table)
	{
		node->prev_node = kmsg_table->prev_node;
		node->next_node = kmsg_table;
		kmsg_table->prev_node->next_node = node;
		kmsg_table->prev_node = node;
	}
	else
	{	
		node->prev_node =
			node->next_node = 
			kmsg_table = node;
	}

	return (node);
}

static void
remove_kmsg_entry (kmsg_timing *node)
{

	if (node)
	{
		if (node->id)
			free (node->id);

		if (node == node->next_node)
		{
			/* If this is the only node */
			kmsg_table = NULL;
		}
		else if (node == kmsg_table)
		{
			/* If this is the root, move the root to the next node */
			kmsg_table = node->next_node;
		}
		node->next_node->prev_node = node->prev_node;
		node->prev_node->next_node = node->next_node;

		free (node);
	}
}

static void
write_timings_msg (kmsg_types type, kmsg_timing *node)
{
	int  elapsed_time;
	char *desc;

	switch (type)
	{
	case WAIT:
		elapsed_time = (node->in_time * 1000 + node->in_millitm) -
								(node->out_time * 1000 + node->out_millitm);
		desc = "WAIT";
		break;
	case PROCESS:
		elapsed_time = (node->out_time * 1000 + node->out_millitm) -
								(node->in_time * 1000 + node->in_millitm);
		desc = "PROCESS";
		break;
	default:
		elapsed_time = 0;
		desc = "UNKNOWN";
		break;
	}

	if (timings_file_name == NULL)
	{
		timings_file_name = malloc (strlen(kqml_agent_name) + 12);
		if (timings_file_name == NULL)
		{
			_kqml_dbg_msg (KQML_DBG_VERBOSE) 
				("Not enough memory for timings file.");
			return;
		}
	}
	sprintf (timings_file_name, "/tmp/%s.Ktime", kqml_agent_name);
	timings_file = fopen (timings_file_name, "a");
	if (timings_file == NULL)
	{
		_kqml_dbg_msg (KQML_DBG_VERBOSE) ("Unable to open KQML timings file %s", 
			timings_file_name);
		return;
	}

	fprintf (timings_file, "%s:%s,%d,%d,%d\n", node->id, 
		desc, node->read_time, node->write_time, elapsed_time);
	fflush (timings_file);
	fclose (timings_file);
}

void
set_kmsg_time (char *id, int timings_type, struct timeval *beg_time, 
	struct timeval *end_time)
{
	kmsg_timing *node;
	long  duration;
  
	/* If timings are not ON, then don't do anything here */
	if ((! kqml_timings ()) || id == NULL)
	{
		return;
	}

	duration = (end_time->tv_sec * 1000 + end_time->tv_usec/1000) -
					(beg_time->tv_sec * 1000 + beg_time->tv_usec/1000);

	/* Find, or create, the entry for this KQML message */
	node = find_kmsg_entry (id);
	if (node == NULL)
	{
		_kqml_dbg_msg (KQML_DBG_VERBOSE) ("Could not add new kmsg node.");
		return;
	}

	/* Set the total time for a READ or WRITE operation */
	if (timings_type == KMSG_READ_TIME)
	{
		node->read_time = duration;
	}
	else
	{
		node->write_time = duration;
	}

	/* the logic for this is a bit hairy, so let me explain; if this is,
		1) a new entry AND a read operation, then this is a message coming 
			from another agent
		2) a new entry AND a write operation, then this is a message 
			originating from the application
		3) an extant entry AND a read operation, then this is an incoming 
			reply to a message sent by this application
		4) an extant entry AND a write operation, then this is an outgoing 
			reply to another agent's message
		A message is new if there is not yet an 'in' time or 'out' time for 
			the message.
		If the message is new, then it's the begining of a new transation. In 
			this case always capture the end time, because the timing is always 
			from the end of the starting operation (whether it's a read or write) 
			to the beginning of the concluding operation (be it a read or a write).
		If the message has been through once already, then this is the 
			concluding transaction, so reverse the above.
  */

	if ((node->in_time == 0) && (node->out_time == 0))
	{
		if (timings_type == KMSG_READ_TIME)
		{
			node->in_time = end_time->tv_sec;
			node->in_millitm = end_time->tv_usec / 1000;
		}
		else
		{
			node->out_time = end_time->tv_sec;
			node->out_millitm = end_time->tv_usec / 1000;
		}
	}
	else
	{
		if (timings_type == KMSG_READ_TIME)
		{
			node->in_time = beg_time->tv_sec;
			node->in_millitm = beg_time->tv_usec / 1000;

			/*	if both 'in' time and 'out' time have been captured, then 
				timings are done */
			if (node->out_time != 0)
			{
				write_timings_msg (WAIT, node);
				remove_kmsg_entry (node);
			}
		}
		else
		{
			node->out_time = beg_time->tv_sec;
			node->out_millitm = beg_time->tv_usec / 1000;

			/* if both 'in' time and 'out' time have been captured, then 
				timings are done */
			if (node->in_time != 0)
			{
				write_timings_msg (PROCESS, node);
				remove_kmsg_entry (node);
			}
		}
	}
}
